from test_treetime import *

test_import_short()

test_assign_gamma()

test_GTR()

test_ancestral()

test_seq_joint_reconstruction_correct()

test_seq_joint_lh_is_max()

print('\n\n TEST HAVE FINISHED SUCCESSFULLY\n\n')
